import matplotlib.pyplot as plt
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report, confusion_matrix
# https://realpython.com/logistic-regression-python/


# For the purpose of this example, let’s just create arrays for the input (𝑥) and output (𝑦) values:
x = np.arange(10).reshape(-1, 1)
y = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])

# The array x is required to be two-dimensional. 

print(x)

# Once you have the input and output prepared, you can create and define your classification model.
model = LogisticRegression(solver='liblinear', random_state=0)
model.fit(x, y)
print(model)

# You can use the fact that .fit() returns the model instance and chain the last two statements.

model = LogisticRegression(solver='liblinear', C=1, random_state=0).fit(x, y)

print(model.intercept_)
#array([-1.04608067])
print(model.coef_,'\n')
# The first column is the probability of the predicted output being zero, that is 1 - 𝑝(𝑥). 
# The second column is the probability that the output is one, or 𝑝(𝑥).
print(model.predict_proba(x),'\n')

print(model.predict(x))
print(model.score(x, y))

# One false positive prediction: The fourth observation is a zero that was wrongly predicted as one.

print(confusion_matrix(y, model.predict(x)))

# It’s often useful to visualize the confusion matrix. 

cm = confusion_matrix(y, model.predict(x))

fig, ax = plt.subplots(figsize=(6, 6))
ax.imshow(cm)
ax.grid(False)
ax.xaxis.set(ticks=(0, 1), ticklabels=('Predicted 0s', 'Predicted 1s'))
ax.yaxis.set(ticks=(0, 1), ticklabels=('Actual 0s', 'Actual 1s'))
ax.set_ylim(1.5, -0.5)
for i in range(2):
    for j in range(2):
        ax.text(j, i, cm[i, j], ha='center', va='center', color='red')
# plt.show()

# The code above creates a heatmap that represents the confusion matrix:
# You can get a more comprehensive report on the classification with 

print(classification_report(y, model.predict(x)))

# You can also implement logistic regression in Python with the StatsModels package. 

import statsmodels.api as sm

y = np.array([0, 1, 0, 0, 1, 1, 1, 1, 1, 1])
x = sm.add_constant(x)
print(x, y)

smodel = sm.Logit(y, x)

result = smodel.fit(method='newton')
print(result.params)
print(result.predict(x))
print((result.predict(x) >= 0.5).astype(int))
print(result.pred_table())
print(result.summary(),'\n')

# These are detailed reports with values that you can obtain with appropriate methods and attributes. 

print(result.summary2())

# For more information, check out the official documentation related to LogitResults.
# https://www.statsmodels.org/stable/generated/statsmodels.discrete.discrete_model.LogitResults.html
